// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.action;

import com.azure.cosmos.FeedOptions;
import com.azure.cosmos.SqlParameter;
import com.azure.cosmos.SqlParameterList;
import com.azure.cosmos.SqlQuerySpec;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.register.action.model.Action;
import org.opengroup.osdu.register.action.persistence.IActionRepo;
import org.opengroup.osdu.azure.CosmosStore;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;
import com.google.cloud.Timestamp;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;

@Repository
public class ActionRepository implements IActionRepo {

    private final ReentrantLock mutex = new ReentrantLock();

    @Autowired
    private String actionContainer;

    @Autowired
    private CosmosStore cosmosStore;

    @Autowired
    private DpsHeaders headers;

    @Autowired
    private String cosmosDBName;

    @Override
    public Action createAction(Action action) {

        if(action.getId() == null){
            throw new AppException(400, "Bad Request", "Action id cannot be null");
        }
        try {
            mutex.lock();

            if (exists(action.getId())) {
                throwConflict();
            }

            ActionDoc doc = new ActionDoc(action, headers.getPartitionId());
            cosmosStore.upsertItem(headers.getPartitionId(), cosmosDBName, actionContainer, doc);

        } finally {
            mutex.unlock();
        }

        return action;
    }

    @Override
    public Action get(String id) {

        ActionDoc doc = cosmosStore.findItem(headers.getPartitionId(), cosmosDBName, actionContainer, id, headers.getPartitionId(), ActionDoc.class)
                .orElseThrow(() -> new AppException(404, "Not found", String.format("Action with id %s does not exist.", id)));

        return convertToActionClass(doc);
    }

    @Override
    public boolean delete(String id) {
        if(!exists(id)){
            return false;
        }

        cosmosStore.deleteItem(headers.getPartitionId(), cosmosDBName, actionContainer, id, headers.getPartitionId());
        return true;
    }

    @Override
    public List<Action> getAllActions() {
        SqlQuerySpec query = new SqlQuerySpec("SELECT * FROM c where c.dataPartitionId = @partitionId");

        SqlParameterList pars = query.getParameters();
        pars.add(new SqlParameter("@partitionId", headers.getPartitionId()));

        FeedOptions options = new FeedOptions();
        return cosmosStore.queryItems(headers.getPartitionId(), cosmosDBName, actionContainer, query, options, ActionDoc.class)
                .stream().map(this::convertToActionClass).collect(Collectors.toList());
    }

    private boolean exists(String id) {
        return cosmosStore.findItem(headers.getPartitionId(), cosmosDBName, actionContainer, id, headers.getPartitionId(), ActionDoc.class).isPresent();
    }

    private Action convertToActionClass(ActionDoc doc){
        Action action = new Action();
        action.setId(doc.getId());
        action.setName(doc.getName());
        action.setDescription(doc.getDescription());
        action.setImg(doc.getImg());
        action.setUrl(doc.getUrl());
        action.setContactEmail(doc.getContactEmail());
        action.setCreatedOnEpoch(Timestamp.of(doc.getCreatedOnEpoch()));
        action.setFilter(doc.getFilter());
        return action;
    }

    private void throwConflict() {
        throw new AppException(409, "Conflict", "An action already exists with the same message and endpoint combination");
    }

}
