package org.opengroup.osdu.register.provider.azure.subscriber;

import com.azure.cosmos.FeedOptions;
import com.azure.cosmos.SqlParameter;
import com.azure.cosmos.SqlParameterList;
import com.azure.cosmos.SqlQuerySpec;
import com.google.cloud.Timestamp;
import org.opengroup.osdu.azure.CosmosStore;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.subscriber.model.*;
import org.opengroup.osdu.register.subscriber.persistence.ISubscriptionRepository;
import org.opengroup.osdu.register.utils.Constants;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.util.List;
import java.util.stream.Collectors;

@Repository
public class SubscriptionRepository implements ISubscriptionRepository {

    @Autowired
    private String subscriptionContainer;

    @Autowired
    private CosmosStore cosmosStore;

    @Autowired
    private DpsHeaders headers;

    @Autowired
    private String cosmosDBName;

    @Override
    public Subscription create(Subscription input) throws Exception {

        // TODO Validate if is there any upper limit required for maximum number of subscriptions for a tenant

       if(exists(input.getId())){
           throw new AppException(409, "Conflict", "A subscriber already exists with the same topic and endpoint combination");
       }

       // TODO Encryption required before storing the secret or Do we need Keyvault to know the secrets

       SubscriptionDoc doc = new SubscriptionDoc(input, headers.getPartitionId());
       cosmosStore.upsertItem(headers.getPartitionId(), cosmosDBName, subscriptionContainer, doc);

       try{
           // TODO Add the flow to create subscription to the given topic

           return input;
       }
       catch (Exception e){
           cosmosStore.deleteItem(headers.getPartitionId(), cosmosDBName, subscriptionContainer, input.getId(), headers.getPartitionId());
           throw new AppException(500, "Server Error", "Unexpected error creating subscription", e);
       }
    }

    @Override
    public Subscription get(String id) {

        SubscriptionDoc doc = cosmosStore.findItem(headers.getPartitionId(), cosmosDBName, subscriptionContainer, id, headers.getPartitionId(), SubscriptionDoc.class)
                .orElseThrow(() -> new AppException(404, "Not found", String.format("Subscriber with id %s does not exist.", id)));

        return convertToSubscriptionClass(doc);
    }

    @Override
    public List<Subscription> query(String notificationId) {

        SqlQuerySpec query = new SqlQuerySpec("SELECT * from c where c.notificationId = @notificationId and c.dataPartitionId = @partitionId");

        SqlParameterList pars = query.getParameters();
        pars.add(new SqlParameter("@notificationId", notificationId));
        pars.add(new SqlParameter("@partitionId", headers.getPartitionId()));

        FeedOptions options = new FeedOptions();

        return cosmosStore.queryItems(headers.getPartitionId(), cosmosDBName, subscriptionContainer, query, options, SubscriptionDoc.class).stream()
                .map(this::convertToSubscriptionClass)
                .collect(Collectors.toList());

    }

    @Override
    public boolean delete(String id) {

        if(!exists(id)){
            return false;
        }

        try{
            // TODO Add flow for deleting the subscription to the topic

            cosmosStore.deleteItem(headers.getPartitionId(), cosmosDBName, subscriptionContainer, id, headers.getPartitionId());
            return true;
        }
        catch (Exception e){
            throw new AppException(500, "Server Error", "Unexpected error deleting subscription.", e);
        }

    }

    @Override
    public boolean patch(Subscription subscription, Secret secret){
        return updateSecret(subscription, secret);
    }

    @Override
    public List<Subscription> getAll() throws Exception {

        SqlQuerySpec query = new SqlQuerySpec("SELECT * from c where c.dataPartitionId = @partitionId");

        SqlParameterList pars = query.getParameters();
        pars.add(new SqlParameter("@partitionId", headers.getPartitionId()));

        FeedOptions options = new FeedOptions();

        return cosmosStore.queryItems(headers.getPartitionId(), cosmosDBName, subscriptionContainer, query, options, SubscriptionDoc.class).stream()
                .map(this::convertToSubscriptionClass)
                .collect(Collectors.toList());
    }

    private boolean updateSecret(Subscription subscription, Secret secret) {
        SubscriptionDoc doc = new SubscriptionDoc(subscription, headers.getPartitionId());
        doc.setSecretType(secret.getSecretType());
        doc.setSecretValue(secret.toString());

        cosmosStore.upsertItem(headers.getPartitionId(), cosmosDBName, subscriptionContainer, doc);
        return true;
    }

    private boolean exists(String id) {
        return cosmosStore.findItem(headers.getPartitionId(), cosmosDBName, subscriptionContainer, id, headers.getPartitionId(), SubscriptionDoc.class).isPresent();
    }

    private Subscription convertToSubscriptionClass(SubscriptionDoc doc) {
        String secretValue = doc.getSecretValue();
        Secret secret;
        if (doc.getSecretType().equals(Constants.GSASecret)) {
            GsaSecret gsaSecret = new GsaSecret();
            String[] splitSecret = secretValue.split("`");
            gsaSecret.setValue(new GsaSecretValue(splitSecret[0], splitSecret[1]));
            secret = gsaSecret;
        } else {
            HmacSecret hmacSecret = new HmacSecret();
            hmacSecret.setValue(secretValue);
            secret = hmacSecret;
        }
        return new Subscription(doc.getId(), doc.getName(),doc.getDescription(), doc.getTopic(), doc.getPushEndpoint(),
                doc.getCreatedBy(), Timestamp.of(doc.getCreatedOnEpoch()), doc.getNotificationId(), secret);
    }
}
