// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.action;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opengroup.osdu.azure.CosmosStore;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.action.model.Action;
import org.opengroup.osdu.register.action.model.Filter;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.lenient;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;

@ExtendWith(MockitoExtension.class)
public class ActionRepositoryTest {

    private static final String dataPartitionId = "data-partition-id";
    private static final String actionId = "id-1";

    @Mock
    private DpsHeaders headers;

    @Mock
    private CosmosStore cosmosStore;

    @InjectMocks
    private ActionRepository repo;

    @BeforeEach
    void init() {
        lenient().doReturn(dataPartitionId).when(headers).getPartitionId();
    }

    @Test
    public void createAction_throws400_whenNullId() {
        Action action = createAction();
        action.setId(null);

        AppException exception = assertThrows(AppException.class, () -> {
            repo.createAction(action);
        });
        assertEquals(400, exception.getError().getCode());
        assertEquals("Bad Request", exception.getError().getReason());
        assertEquals("Action id cannot be null", exception.getError().getMessage());
    }

    @Test
    public void createAction_throws409_whenIdExists() {
        Action action = createAction();
        Optional<ActionDoc> cosmosItem = Optional.of(new ActionDoc());

        doReturn(cosmosItem).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        AppException exception = assertThrows(AppException.class, () -> {
            repo.createAction(action);
        });

        assertEquals(409, exception.getError().getCode());
        assertEquals("Conflict", exception.getError().getReason());
        assertEquals("An action already exists with the same message and endpoint combination", exception.getError().getMessage());
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, actionId, dataPartitionId, ActionDoc.class);
    }

    @Test
    public void createAction_documentInsertedSuccessfully() {
        Action action = createAction();

        Optional<ActionDoc> cosmosItem = Optional.empty();
        doReturn(cosmosItem).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());
        doNothing().when(cosmosStore).upsertItem(anyString(), any(), any(), any());

        Action output = repo.createAction(action);

        assertEquals(action, output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, actionId, dataPartitionId, ActionDoc.class);
    }

    @Test
    public void getAction_throws404_whenIdNotFound() {

        doReturn(Optional.empty()).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        AppException exception = assertThrows(AppException.class, () -> {
            repo.get(actionId);
        });

        assertEquals(404, exception.getError().getCode());
        assertEquals("Not found", exception.getError().getReason());
        assertEquals("Action with id id-1 does not exist.", exception.getError().getMessage());
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, actionId, dataPartitionId, ActionDoc.class);
    }

    @Test
    public void getAction_itemRetrievedSuccessfully() {

        Action action = createAction();
        ActionDoc doc = new ActionDoc(action, dataPartitionId);

        doReturn(Optional.of(doc)).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        Action output = repo.get(actionId);

        assertEquals(action, output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, actionId, dataPartitionId, ActionDoc.class);
    }

    @Test
    public void deleteAction_returnsFalse_whenIdNotFound() {

        doReturn(Optional.empty()).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        boolean output = repo.delete(actionId);

        assertFalse(output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, actionId, dataPartitionId, ActionDoc.class);
    }

    @Test
    public void deleteAction_returnsTrue_whenDeletedSuccessfully() {

        Action action = createAction();
        ActionDoc doc = new ActionDoc(action, dataPartitionId);

        Optional<ActionDoc> cosmosItem = Optional.of(doc);
        doReturn(cosmosItem).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        boolean output = repo.delete(actionId);

        assertTrue(output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, actionId, dataPartitionId, ActionDoc.class);
        verify(cosmosStore, times(1)).deleteItem(dataPartitionId, null, null, actionId, dataPartitionId);
    }

    @Test
    public void getAllActions_returnAllItemsSuccessfully() {

        Action action1 = createAction();
        Action action2 = createAction();
        action2.setId("id-2");

        ActionDoc doc1 = new ActionDoc(action1, dataPartitionId);
        ActionDoc doc2 = new ActionDoc(action2, dataPartitionId);

        List<Action> actions = Stream.of(action1, action2).collect(Collectors.toList());
        List<ActionDoc> actionDocs = Stream.of(doc1, doc2).collect(Collectors.toList());

        doReturn(actionDocs).when(cosmosStore).queryItems(anyString(), any(), any(), any(), any(), any());

        List<Action> output = repo.getAllActions();

        actions.get(0).setCreatedOnEpoch(output.get(0).getCreatedOnEpoch());
        actions.get(1).setCreatedOnEpoch(output.get(1).getCreatedOnEpoch());

        assertEquals(actions, output);
    }

    private Action createAction() {
        Action action = new Action();
        action.setId(actionId);
        action.setName("name");
        action.setDescription("description");
        action.setImg("https://mycdn.com/img.png");
        action.setUrl("https://myapp.osdu.opengroup.org/action/{id}/{data.project}");
        action.setFilter(new Filter());

        return action;
    }
}