// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.ddms;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opengroup.osdu.azure.CosmosStore;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.ddms.model.Ddms;
import org.opengroup.osdu.register.ddms.model.RegisteredInterface;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.lenient;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;

@ExtendWith(MockitoExtension.class)
public class DdmsRepositoryTest {

    private static final String dataPartitionId = "data-partition-id";
    private static final String ddmsId = "id-1";

    @Mock
    private DpsHeaders headers;

    @Mock
    private CosmosStore cosmosStore;

    @InjectMocks
    private DdmsRepository repo;

    @BeforeEach
    void init() {
        lenient().doReturn(dataPartitionId).when(headers).getPartitionId();
    }

    @Test
    public void createDdms_throws409_whenIdExists() {
        Ddms ddms = createDdms();
        Optional<DdmsDoc> cosmosItem = Optional.of(new DdmsDoc());

        doReturn(cosmosItem).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        AppException exception = assertThrows(AppException.class, () -> {
            repo.create(ddms);
        });

        assertEquals(409, exception.getError().getCode());
        assertEquals("Conflict", exception.getError().getReason());
        assertEquals("A DDMS already exists with the same id", exception.getError().getMessage());
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, ddmsId, dataPartitionId, DdmsDoc.class);
    }

    @Test
    public void createDdms_documentInsertedSuccessfully() {
        Ddms ddms = createDdms();

        Optional<DdmsDoc> cosmosItem = Optional.empty();
        doReturn(cosmosItem).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());
        doNothing().when(cosmosStore).upsertItem(anyString(), any(), any(), any());

        Ddms output = repo.create(ddms);

        assertEquals(ddms, output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, ddmsId, dataPartitionId, DdmsDoc.class);
    }

    @Test
    public void getDdms_throws404_whenIdNotFound() {

        doReturn(Optional.empty()).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        AppException exception = assertThrows(AppException.class, () -> {
            repo.get(ddmsId);
        });

        assertEquals(404, exception.getError().getCode());
        assertEquals("Not found", exception.getError().getReason());
        assertEquals("DDMS with id id-1 does not exist.", exception.getError().getMessage());
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, ddmsId, dataPartitionId, DdmsDoc.class);
    }

    @Test
    public void getDdms_itemRetrievedSuccessfully() {

        Ddms ddms = createDdms();
        DdmsDoc doc = new DdmsDoc(ddms, dataPartitionId);

        doReturn(Optional.of(doc)).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        Ddms output = repo.get(ddmsId);

        assertEquals(ddms, output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, ddmsId, dataPartitionId, DdmsDoc.class);
    }

    @Test
    public void deleteDdms_returnsFalse_whenIdNotFound() {

        doReturn(Optional.empty()).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        boolean output = repo.delete(ddmsId);

        assertFalse(output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, ddmsId, dataPartitionId, DdmsDoc.class);
    }

    @Test
    public void deleteDdms_returnsTrue_whenDeletedSuccessfully() {

        Ddms ddms = createDdms();
        DdmsDoc doc = new DdmsDoc(ddms, dataPartitionId);

        Optional<DdmsDoc> cosmosItem = Optional.of(doc);
        doReturn(cosmosItem).when(cosmosStore).findItem(anyString(), any(), any(), anyString(), anyString(), any());

        boolean output = repo.delete(ddmsId);

        assertTrue(output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, null, null, ddmsId, dataPartitionId, DdmsDoc.class);
        verify(cosmosStore, times(1)).deleteItem(dataPartitionId, null, null, ddmsId, dataPartitionId);
    }

    @Test
    public void queryDdms_returnItemsSuccessfully() {

        Ddms ddms1 = createDdms();
        Ddms ddms2 = createDdms();
        ddms2.setId("id-2");

        DdmsDoc doc1 = new DdmsDoc(ddms1, dataPartitionId);
        DdmsDoc doc2 = new DdmsDoc(ddms2, dataPartitionId);

        List<Ddms> ddmsList = Stream.of(ddms1, ddms2).collect(Collectors.toList());
        List<DdmsDoc> ddmsDocList = Stream.of(doc1, doc2).collect(Collectors.toList());

        doReturn(ddmsDocList).when(cosmosStore).queryItems(anyString(), any(), any(), any(), any(), any());

        List<Ddms> output = repo.query("type");

        ddmsList.get(0).setCreatedDateTimeEpoch(output.get(0).getCreatedDateTimeEpoch());
        ddmsList.get(1).setCreatedDateTimeEpoch(output.get(1).getCreatedDateTimeEpoch());

        assertEquals(ddmsList, output);
    }

    private Ddms createDdms() {
        Ddms ddms = new Ddms();
        RegisteredInterface ri = new RegisteredInterface();
        ri.setEntityType("type");
        ri.setSchema(Collections.singletonMap("first", "second"));

        ddms.setId(ddmsId);
        ddms.setName("mock-name");
        ddms.setDescription("mock-description");
        ddms.setContactEmail("mock-contactEmail");
        ddms.getInterfaces().add(ri);
        return ddms;
    }

}
